//FRAGMENT SHADER
#version 430 core

//Constants
const float PI = 3.1415926535897932384626433832795;

//Settings
#define POINT_LIGHT_COUNT 6
#define DIR_LIGHT_COUNT 6
#define SPOT_LIGHT_COUNT 6
#define MAX_SHADOW_MAPPING_KERNEL_SIZE 5

// Lights & their properties
// Point Light
struct PointLight {
	vec3 position;
	vec3 intensity;
	vec3 attenuation;
};
uniform PointLight pointLights[POINT_LIGHT_COUNT];

// Directional Light
struct DirectionalLight {
	vec3 direction;
	vec3 intensity;
};
uniform DirectionalLight directionalLights[DIR_LIGHT_COUNT];

// SpotLight
struct SpotLight {
	vec3 position;
	vec3 direction;
	vec3 intensity;
	vec3 attenuation;
	float outerAngle;
	float innerAngle;
};
uniform SpotLight spotLights[SPOT_LIGHT_COUNT];

// Camera info
uniform vec3 eyePos;

// Shadow Mapping
in vec4 fragPositionInLightSpace;
uniform uvec2 shadowMapKernelSize;
uniform sampler2D shadowMap;
uniform vec3 shadowSourceLightDirection;
uniform float[MAX_SHADOW_MAPPING_KERNEL_SIZE * MAX_SHADOW_MAPPING_KERNEL_SIZE] shadowMapKernel;

// Material properties
uniform vec3 albedo;
uniform float ao;
uniform float roughness;
uniform float metalness;

in vec3 fragPosition;
in vec3 fragNormal;

layout (location = 0) out vec4 fragColor;
layout (location = 1) out vec4 brightFragColor;

// =====================
// Function declarations
// =====================

// Light source radiances
vec3 getDirectionalLightRadiance(DirectionalLight light, vec3 n, vec3 v);
vec3 getPointLightRadiance(PointLight light, vec3 n, vec3 v);
vec3 getSpotLightRadiance(SpotLight light, vec3 n, vec3 v);

// Physically Based Rendering - derived from https://learnopengl.com/PBR/Theory

// Trowbridge-Reitz GGX Normal Distribution Function
// Input:
//	- NdotH: precomputed max(dot(n, h), 0.0F), [0, 1]
//	- alpha: Roughness, [0, 1]
float NDF_GGXTR(float NdotH, float alpha);

// Schlick-GGX Geometry Function
// Input:
//  - NdotX: precomputed max(dot(n, v or l), 0.0F), [0, 1]
//  - k: Remapping of alpha (roughness)
float G_SchlickGGX(float NdotX, float k);

// Smiths method to take into account "geometry obstruction" and "geometry shadowing".
// Input:
//  - NdotV: precomputed max(dot(n, v), 0.0F), [0, 1]
//  - NdotL: precomputed max(dot(n, l), 0.0F), [0, 1]
//  - alpha: Roughness, [0, 1]
float G_Smith(float NdotV, float NdotL, float alpha);

// Fresnel-Schlick approximation
// Input:
//  - HdotV: precomputed max(dot(h, v), 0.0F), [0, 1]
vec3 F_Schlick(float HdotV);

// Cook-Torrance BRDF
// Input:
//  - NdotV: precomputed max(dot(n, v), 0.0F), [0, 1]
//  - NdotL: precomputed max(dot(n, l), 0.0F), [0, 1]
//	- NdotH: precomputed max(dot(n, h), 0.0F), [0, 1]
//  - HdotV: precomputed max(dot(h, v), 0.0F), [0, 1]
//  - F: precomputed Fresnel-Schlick for the current light,  [0, 1]
vec3 BRDF_CookTorrance(float NdotV, float NdotL, float NdotH, float HdotV, vec3 F);

// Shadow Mapping - derived from https://learnopengl.com/Advanced-Lighting/Shadows/Shadow-Mapping
float calculateShadow(vec3 n);

// =====================
// == Global variables =
// =====================
vec3 F0 = vec3(0.04F); // TODO: add description


// =====================
// ======= Main ========
// =====================
void main() {
	vec3 n = normalize(fragNormal);
	vec3 v = normalize(eyePos - fragPosition);

	// Irradiance = The Sum of all radiances
	vec3 irradiance = vec3(0.0F);

	// Precompute Base reflectivity for use on dielectrics and conductors
	F0 = mix(F0, albedo, metalness);

	// Add radiances of point lights to irradiance
	for (int i = 0; i < POINT_LIGHT_COUNT; i++) {
		irradiance += getPointLightRadiance(pointLights[i], n, v);
	}

	// Add radiances of directional lights to irradiance
	for (int i = 0; i < DIR_LIGHT_COUNT; i++) {
		irradiance += getDirectionalLightRadiance(directionalLights[i], n, v);
	}

	// Add radiances of spot lights to irradiance
	for (int i = 0; i < SPOT_LIGHT_COUNT; i++) {
		irradiance += getSpotLightRadiance(spotLights[i], n, v);
	}
	
	// Calculate shadow value and adjust irradiance
	float shadow = calculateShadow(n);
	// TODO: apply only to the shadowing light source?
	irradiance *= vec3(1-shadow/2); // TODO: shadow intensity

	// Add ambient color
	irradiance += vec3(0.03F) * albedo * ao;

	// Tone mapping
	// irradiance = irradiance / (irradiance + vec3(1.0F));

	// Gamma correction
	// irradiance = pow(irradiance, vec3(1.0F / 2.2F));

	fragColor = vec4(irradiance, 1.0F);
	
	// Set fragment in bright texture for use in bloom
	float brightness = dot(irradiance.rgb, vec3(0.2126, 0.7152, 0.0722));
    if(brightness > 0.3) // TODO fine tune
        brightFragColor = vec4(fragColor.rgb, 1.0);
    else
        brightFragColor = vec4(0.0, 0.0, 0.0, 1.0);
	//fragColor = vec4(fragNormal, 1.0F);
}

vec3 getDirectionalLightRadiance(DirectionalLight light, vec3 n, vec3 v) {
	if (light.intensity == vec3(0.0F)) {
		//Skip light if intensity is zero
		return vec3(0.0F);
	} 

	// To Light vector
	vec3 l = normalize(-light.direction);
	
	// Normalized halfway vector between light vector and view vector
	vec3 h = normalize(v + l);

	// Precompute dot products for the BRDF
	float NdotV = max(dot(n, v), 0.0F);
	float NdotL = max(dot(n, l), 0.0F);
	float NdotH = max(dot(n, h), 0.0F);
	float HdotV = max(dot(h, v), 0.0F);

	// Precompute Fresnel-Schlick
	vec3 F = F_Schlick(HdotV);

	// Calculate specular/diffuse ratio for fragment
	vec3 kS = F;
	vec3 kD = vec3(1.0F) - kS;

	// Diminish diffuse reflection for metallic objects
	kD *= 1.0F - metalness;

	vec3 cookTorranceSpecular = BRDF_CookTorrance(NdotV, NdotL, NdotH, HdotV, F);

	return (kD * albedo / PI + cookTorranceSpecular) * light.intensity * NdotL;
}

vec3 getPointLightRadiance(PointLight light, vec3 n, vec3 v) {
	if (light.intensity == vec3(0.0F)) {
		//Skip light if intensity is zero
		return vec3(0.0F);
	} 

	// To Light vector
	vec3 l = light.position - fragPosition;
	
	// Distance between fragment <-> light
	float d = length(l);
	
	// Normalize to light vector
	l = normalize(l);
	
	// Normalized halfway vector between light vector and view vector
	vec3 h = normalize(v + l);

	// Attenuation through the inverse square law
	float attenuation = 1.0F / (d * d);

	// Precompute dot products for the BRDF
	float NdotV = max(dot(n, v), 0.0F);
	float NdotL = max(dot(n, l), 0.0F);
	float NdotH = max(dot(n, h), 0.0F);
	float HdotV = max(dot(h, v), 0.0F);

	// Precompute Fresnel-Schlick
	vec3 F = F_Schlick(HdotV);

	// Calculate specular/diffuse ratio for fragment
	vec3 kS = F;
	vec3 kD = vec3(1.0F) - kS;

	// Diminish diffuse reflection for metallic objects
	kD *= 1.0F - metalness;

	vec3 cookTorranceSpecular = BRDF_CookTorrance(NdotV, NdotL, NdotH, HdotV, F);

	return (kD * albedo / PI + cookTorranceSpecular) * light.intensity * attenuation * NdotL;
}

vec3 getSpotLightRadiance(SpotLight light, vec3 n, vec3 v) {
	if (light.intensity == vec3(0.0F)) {
		//Skip light if intensity is zero
		return vec3(0.0F);
	} 

	// To Light vector
	vec3 l = light.position - fragPosition;
	
	// Distance between fragment <-> light
	float d = length(l);
	
	// Normalize to light vector
	l = normalize(l);
	
	// Normalized halfway vector between light vector and view vector
	vec3 h = normalize(v + l);

	// Attenuation through the inverse square law
	float attenuation = 1.0F / (d * d);

	// Calculate how similar the light vector is to the light direction
	float LdotLightDir = max(dot(l, normalize(-light.direction)), 0.0F);

	// Calculate the attenuation based on the cone in which the light shines.
	attenuation *= smoothstep(light.innerAngle, light.outerAngle, LdotLightDir);

	// Precompute dot products for the BRDF
	float NdotV = max(dot(n, v), 0.0F);
	float NdotL = max(dot(n, l), 0.0F);
	float NdotH = max(dot(n, h), 0.0F);
	float HdotV = max(dot(h, v), 0.0F);

	// Precompute Fresnel-Schlick
	vec3 F = F_Schlick(HdotV);

	// Calculate specular/diffuse ratio for fragment
	vec3 kS = F;
	vec3 kD = vec3(1.0F) - kS;

	// Diminish diffuse reflection for metallic objects
	kD *= 1.0F - metalness;

	vec3 cookTorranceSpecular = BRDF_CookTorrance(NdotV, NdotL, NdotH, HdotV, F);

	return (kD * albedo / PI + cookTorranceSpecular) * light.intensity * attenuation * NdotL;
}

float NDF_GGXTR(float NdotH, float alpha) {
	float alpha_2 = alpha * alpha;
	float NdotH_2 = NdotH * NdotH;

	float denominator = NdotH_2 * (alpha_2 - 1.0F) + 1.0F;
	denominator = PI * denominator * denominator;

	return alpha_2 / max(denominator, 0.0000001F); // Prevent div by 0 error
}

float G_SchlickGGX(float NdotX, float k) {
	float numerator = NdotX;
	float denominator = NdotX * (1.0F - k) + k;

	return numerator / max(denominator, 0.0000001F); // Prevent div by 0 error
}

float G_Smith(float NdotV, float NdotL, float alpha) {
	//Remapping roughness
	float k_numerator = alpha + 1;
	float k = (k_numerator * k_numerator) / 8.0F;

	//Calculate "geometry obstruction" times "geometry shadowing"
	return G_SchlickGGX(NdotV, k) * G_SchlickGGX(NdotL, k);
}

vec3 F_Schlick(float HdotV) {
	return F0 + (1.0F - F0) * pow(1.0F - HdotV, 5.0F);
}

vec3 BRDF_CookTorrance(float NdotV, float NdotL, float NdotH, float HdotV, vec3 F) {
	float NDF = NDF_GGXTR(NdotH, roughness);
	float G = G_Smith(NdotV, NdotL, roughness);

	vec3 numerator = NDF * G * F;
	float denominator = 4.0F * NdotV * NdotL;

	return numerator / max(denominator, 0.0000001F); // Prevent div by 0 error
}

float calculateShadow(vec3 n) {
	// if we are looking away from the light we are in shadow.
	if (dot(n, shadowSourceLightDirection) > 0) {
		return 1;
	}

	float shadow = 0.0;
	vec2 texelSize = 1.0 / textureSize(shadowMap, 0);

	// normalize the fragPosition in light space (needed for perspective light view matrices) and map to [0, 1]
	vec3 projectedCoords = fragPositionInLightSpace.xyz / fragPositionInLightSpace.w; // [-1, 1]
	projectedCoords = projectedCoords * 0.5 + 0.5; // [0, 1]

	// we are farther away than the far plane or closer than the near plane (of the light view)
	if(projectedCoords.z >= 1.0) {
		return 0.0;
	}

	// bias towards being lit (otherwise due to depth sampling we see shadows in lit areas)
	float bias = max(1.0 - dot(n, shadowSourceLightDirection), 1.0) * 0.00005;

	// depth from the cameras point of view (to the current fragment)
	float depthForCamera = projectedCoords.z;

	ivec2 shadowMapKernelExtends = (ivec2(shadowMapKernelSize.x, shadowMapKernelSize.y)-1) / 2;

	uint a = 0;
	for(int kernelX = -shadowMapKernelExtends.x; kernelX <= shadowMapKernelExtends.x; kernelX++) {
		for(int kernelY = -shadowMapKernelExtends.y; kernelY <= shadowMapKernelExtends.y; kernelY++) {
			// depth from the lights point of view (to the current fragment)
			float depthForLight = texture(shadowMap, projectedCoords.xy + vec2(kernelX, kernelY) * texelSize).r; 

			// shadow value for this texel [0,1]
			float shadowForTexel = (depthForCamera - bias) > depthForLight ? 1.0 : 0.0;
			
			// weight for the shadow value of the current texel - prioritizes texels closer to the center
			float texelWeight = shadowMapKernel[(kernelX+shadowMapKernelExtends.x) + (kernelY+shadowMapKernelExtends.y) * shadowMapKernelSize.y];

			// add to accumulators
			shadow += shadowForTexel * texelWeight;
		}    
	}

	return shadow;
}