//VERTEX SHADER
#version 430 core

#define MAX_BONE_TRANSFORMS 50
#define MAX_BONES_PER_VERTEX 4 // This is equal to the size of the vectors influencingBones and boneWeights below.

layout(location = 0) in vec3 position;
layout(location = 1) in vec3 normal;
layout(location = 3) in uvec4 influencingBones;
layout(location = 4) in vec4 boneWeights;

uniform bool hasBones;

uniform mat4 boneTransforms[MAX_BONE_TRANSFORMS];

uniform mat4 modelMatrix;
uniform mat4 viewProjectMatrix;

// shadow mapping
out vec4 fragPositionInLightSpace;
uniform mat4 shadowSourceLightSpaceMatrix;

out vec3 fragPosition;
out vec3 fragNormal;

void main() {

	vec4 modelSpaceVertexTransformAccum = vec4(0.0F);
	vec4 modelSpaceVertexNormalAccum = vec4(0.0F);
	
	if (hasBones) {
		for (int i = 0; i < MAX_BONES_PER_VERTEX; i++) {
			// Get the tranform of the current bone
			mat4 boneTransform = boneTransforms[influencingBones[i]];

			// Transform the current vertex position according to the current bone transform
			vec4 posTransformedByCurrentBone = boneTransform * vec4(position, 1.0);

			// Weight the position transformation of the current bone by multiplying with the bone's weight
			// and add it to the vertex position accumulator in model space
			modelSpaceVertexTransformAccum += posTransformedByCurrentBone * boneWeights[i];

			// Transform the current vertex normal according to the current bone transform
			vec4 normalTransformedByCurrentBone = boneTransform * vec4(normal, 0.0);

			// Weight the normal transformation of the current bone by multiplying with the bone's weight
			// and add it to the vertex normal accumulator in model space
			modelSpaceVertexNormalAccum += normalTransformedByCurrentBone * boneWeights[i];
		}
	} else {
		modelSpaceVertexTransformAccum = vec4(position, 1.0F);
		modelSpaceVertexNormalAccum = vec4(normal, 0.0F);
	}

	// Store the Vertex position in gl_Position (for further fragment shader use) with the homogenous coord w = 1.
	vec4 worldPosition = modelMatrix * modelSpaceVertexTransformAccum;

	//Apply the viewProject matrix to the point to calculate the screen coordinates and pass them to OpenGL.
	gl_Position = viewProjectMatrix * worldPosition;
	
	//Transfer the worldPosition to the fragment shader
	fragPosition = worldPosition.xyz / worldPosition.w;
	//Calculate the transformed normal with the transposed inverse of the model matrix and send it to the fragment shader.
	fragNormal = vec3(transpose(inverse(modelMatrix)) * modelSpaceVertexNormalAccum);

	fragPositionInLightSpace = shadowSourceLightSpaceMatrix * vec4(fragPosition, 1.0);
}
